/* N-body calculation using Greenspan 1st-order discrete mechanics.	*/
/* Uses N*N*Q processors, communication by spreading and reduction,	*/
/* explict loops for xyz.  All processors store particle information	*/
/* and calculate velocities, positions.					*/
/* Uses polystep technique, integrating Q time steps in parallel.	*/
/* Uses dynamic shape allocation for flexibility.			*/
/* Uses some paris code for spreading.					*/
/***********  Production system with checkpointing on datavault *********/
/* Checkpoint files are exactly the same as for nbpolyp.cs, 		*/
/* ***** Rx must be maintained separately *****				*/
/* Provides for graphical output using solplot.cs.			*/
/* Incorporates repulsive force to handle asteroid collisions using	*/
/* force  H/r^8  ,  H is chosen so that force is 0 at r = Rx.		*/
/* File:  axpoly.cs							*/

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <cscomm.h>
#include <cm/paris.h>
#include <cstimer.h>
#include <cm/cmfs.h>
#include <cm/cm_file.h>
#include "colorp.h"
#include "solplot.h"

/* Max number of bodies in the calculation.  Q = 4, N = 64 => 16k procs */
/* Q, N must be powers of 2, Q*N^2 >= physical CM size.			*/

int Q =  4;		/* Number of parallel time steps.		*/
int Q0 = 0;		/* Index of block with current time step	*/
int N = 64;		/* Max number of bodies permitted		*/
int   NB;		/* Number of bodies in system, <= N		*/
int   PX;		/* Controls how much state is printed		*/
int PD = 0;		/* Controls diagnostic printing (0 = off)	*/
float G;		/* Gravitational constant			*/
float H, HG;		/* Asteroid repulsion constant, H/G		*/
float Rx=0.00005;	/* Radius for asteroid repulsion		*/

int advance;	/* Indicates that time step should be advanced.		*/

#define TWOPI      6.2831853f

/* Particles, and pairs of particles.  All entries have particle	*/
/* states.  Some variables could be overloaded. 			*/

shape [][][]pair;	/* Will be [Q][N][N]	*/

/* All parallel variables are declared as pointers so 'pair' can be dynamic.   */
/* State variables will be allocated as x0[3], etc.;  w, m, mj, r0, are floats */

/* x0, x1 - current, new positions;	v0, v1 - current, new velocities */
/* m - mass;		mj - scaled mass;	ac - acceleration	 */
/* w, ax - working	r0 - distance (X0)	d0 - displacement (X0)	 */

float:pair *x0, *x1, *v0, *v1, *m, *mj, *ac, *d0, *r0, *ax, *w;

/* i - row number;    j - column number;    c - block number	*/

int:pair *i, *j, *c;

/* sm   - processor is part of computation = diag | act	*/
/* diag - processor is on diagonal			*/
/* act  - processor is active pair (off-diag)		*/
/* sp   - block calculating current time step		*/

bool:pair *sm, *diag, *act, *sp;

/* CM addresses for Paris code */

CM_memaddr_t diag_addr, ax_addr, x1_addr, x0_addr, v1_addr, v0_addr, d0_addr;

/* Allocate global parallel variables, pair must have been allocated.	*/
/* Sizes used elsewhere are global.					*/

int fsz, vsz;	/* Sizes of float, float 3-vector */

void create_globals()
 { int isz, bsz;
   fsz = boolsizeof(float:pair);  vsz = 3*fsz;
   isz = boolsizeof(int:pair);
   bsz = boolsizeof(bool:pair);
   /* Float vectors[3] */
   x0 = palloc(pair, vsz);  x1 = palloc(pair, vsz);
   v0 = palloc(pair, vsz);  v1 = palloc(pair, vsz);
   ac = palloc(pair, vsz);  ax = palloc(pair, vsz);
   d0 = palloc(pair, vsz);
   /* Floats */
   m  = palloc(pair, fsz);  mj = palloc(pair, fsz);
   r0 = palloc(pair, fsz);  w  = palloc(pair, fsz);
   /* Ints and bools */
   i  = palloc(pair, isz);  j = palloc(pair, isz);  c = palloc(pair, isz);
   sm  = palloc(pair, bsz);  diag = palloc(pair, bsz);
   act = palloc(pair, bsz);  sp   = palloc(pair, bsz);
   /* Fill in addresses for Paris */
   diag_addr = (CM_memaddr_t)diag;  ax_addr = (CM_memaddr_t)ax;
   x1_addr = (CM_memaddr_t)x1;      x0_addr = (CM_memaddr_t)x0;
   v1_addr = (CM_memaddr_t)v1;      v0_addr = (CM_memaddr_t)v0;
   d0_addr = (CM_memaddr_t)d0;
 }



/* Pairs are represented symmetrically.	 Setup initializes these.	*/

void setup()
 { with (pair)
    { /* Initialize coordinates */
      *c = pcoord(0);  *i = pcoord(1);  *j = pcoord(2);
      /* Identify active processors */
      *sm = (*i < NB) & (*j < NB);
      *act = *sm & (*i != *j);  *diag = *sm & (*i == *j);
    }
 }

FILE *datap;      /* Stream for reading data */

/* Load inital positions, velocities, also masses, gravitational const. */
/* Leaves data for Pi in processors [0][i][*].				*/

void init()
 { int k;  float ax, ay, az;
   Q0 = 0;	/* Block 0 will be initial time step */
   with (pair)
   {*sp = (*c == Q0);
    for ( k = 0; k < NB; k++ )
     { fscanf(datap, "%e%e%e", &ax, &ay, &az);
       [0][k][0]x0[0] = ax;  [0][k][0]x0[1] = ay;  [0][k][0]x0[2] = az;
     };
    for ( k = 0; k < NB; k++ )
     { fscanf(datap, "%e%e%e", &ax, &ay, &az);
       [0][k][0]v0[0] = ax;  [0][k][0]v0[1] = ay;  [0][k][0]v0[2] = az;
     };
    for ( k = 0; k < NB; k++ )
      { fscanf(datap, "%e", &ax);  [0][k][0]*m = ax; };
    fscanf(datap, "%e", &G);
    /* Spread data across rows */
    x0[0] = copy_spread(&x0[0], 2, 0);
    x0[1] = copy_spread(&x0[1], 2, 0);
    x0[2] = copy_spread(&x0[2], 2, 0);
    v0[0] = copy_spread(&v0[0], 2, 0);
    v0[1] = copy_spread(&v0[1], 2, 0);
    v0[2] = copy_spread(&v0[2], 2, 0);
    *m  = copy_spread(m,  2, 0);
   }
 }

/* Insert nx asteroids at radius ar with total mass am.  This assumes	*/
/* that an orbit of radius 1 about origin corresponds to velocity 2*pi	*/
/* (e.g., au-earth-yr units).  Uses parallel trig.			*/

void mkast(int nx, float ar, float am)
 { float:pair sx, cx, arg;
   float spd;  int nn;
   nn = NB + nx;
   spd = TWOPI/sqrt(ar);
   printf("Making %d asteroids, r = %9.5f, spd = %9.5f, tot mass = %9.2e\n",
		  nx,		     ar,	  spd,		    am);
   with (pair) 
   { *sm = (*i < nn) & (*j < nn);
     *act = *sm & (*i != *j);  *diag = *sm & (*i == *j);
     where ( (NB <= *i) & (*i < nn) & *sm )
      { arg = (TWOPI*(*i - NB))/nx;
        sx = sin(arg);  cx = cos(arg);
        x0[0] =  cx*ar;  v0[0] = (-sx)*spd;
        x0[1] =  sx*ar;  v0[1] =  cx*spd;
        x0[2] = 0.0;     v0[2] = 0.0;
        *m = am/nx;
      }}  /* end where, end with (pair) */
   NB = nn;
 }

/* Print current state of system according to PX.  See comments for	*/
/* main program for PX settings.  Takes `advance' into account in	*/
/* determining which values to print.					*/
/* Must be called with pair selected when plotting.			*/

void print_state()
 { int k, np;  float:current *x, *v;
   if ( (PX >= 0) && (PX < NB) ) np = PX; else np = NB;
   if ( advance ) { x = x1;  v = v1; } else { x = x0;  v = v0; }
   if ( PX == -2 ) { where ( *c == Q0 ) plot_planets(x); }
   else
   { printf(
  "\n       x         y         z         vx        vy        vz        m\n");
   for ( k = 0; k < np; k++ )
     printf("%2d %9.5f %9.5f %9.5f %9.5f %9.5f %9.5f %9.2e\n",
            k, [Q0][k][k]x[0], [Q0][k][k]x[1], [Q0][k][k]x[2],
               [Q0][k][k]v[0], [Q0][k][k]v[1], [Q0][k][k]v[2],
	       [Q0][k][k]*m					 );
   }
 }

/* Print x-components of xx in nx by nx processors (for testing). */
/* Values printed from time slice q >= 0 (typically Q0)		  */
/* nx = 7 fits nicely on one line.				  */
/*******************************************************************/
void print_x(char lbl[], float:current *xx, int nx, int q)
 { int u, v;
   if ( q >= Q ) q = q%%Q;	/* Keep in range */
   if (PD)
    { printf("\n%s  -  slice %d\n", lbl, q); 
      for ( u = 0; u < nx; u++)
       { for ( v = 0;  v < nx; v++)
           printf("%9.2e ", [q][u][v]*xx);
         printf("\n");
       }
    }
 }
/********************************************************************/

/* Print x-components of xx in nx by nx processors (for testing). */
/* Values printed from time slice q >= 0 (typically Q0)		  */
/******************************************************************
void print_b(char lbl[], bool:current *xx, int nx, int q)
 { int u, v;
   if ( q >= Q ) q = q%%Q;	*//* Keep in range *//*
   if (PD)
    { printf("\n%s  -  slice %d\n", lbl, q); 
      for ( u = 0; u < nx; u++)
       { for ( v = 0;  v < nx; v++)
           printf("%1d ", [q][u][v]*xx);
         printf("\n");
       }
    }
 }
*******************************************************************/

/* Spread xv[3] along i-axis (1-axis) from diagonal, leaving result in dv, */
/* Called with all processors active and leaves all active.  Uses ax.	   */
/* 'vsz' is size of float 3-vector, defined above.			   */

void spreadi(float:pair *dv, float:pair *xv)
 { CM_memaddr_t dv_addr, xv_addr;
   dv_addr = dv;  xv_addr = xv;				/*  For types 	*/
   CM_u_move_zero_always_1L(ax_addr, vsz);		/* wv := 0	*/
   CM_load_context(diag_addr);				/* where (diag)	*/
   CM_u_move_1L(ax_addr, xv_addr, vsz); 		/*    wv := xv	*/
   CM_set_context();					/* all active	*/
   CM_spread_with_logior_1L(dv_addr, ax_addr, 1, vsz);	/* Spread	*/
   /* if (PD) printf("Leaving spreadi\n");		For testing	*/
 }

/* Calculate accelerations.  r0, d0 already give displacement, distance  */
/* for current position.  Called with all processors active.		 */

void accel()
 { float:pair d1[3], r1, p0, p1, r01;  int u;  CM_memaddr_t d1_addr;
   d1_addr = &d1[0];
   *w = p1 = 0.0f;		/* To accumulate sum of squares (later) */
   for ( u = 0;  u < 3;  u++ ) ac[u] = 0.0f;
   spreadi(&d1[0], x1);		/* Spread x1 to d1 */
   /* Now d1 = [c][j][j]x1,  x1 = [c][i][i]x1 */
   where (*act) 
    { for ( u = 0;  u < 3;  u++ )
       { r1 = d1[u] -= x1[u];		/* New displacement component */
         p1 += r1*r1;			/* For sum of squares */
       }
      r1 = sqrt(p1);		/* Euclidean distance */
    } /* end where (*act) ... */
   if (advance)		/* Advance time step	*/
    { Q0++;  if ( Q0 == Q ) Q0 = 0;  advance = 0;
      to_torus_dim(x0, x1, vsz, 0, 1);  to_torus_dim(v0, v1, vsz, 0, 1);
      to_torus_dim(d0, d1, vsz, 0, 1);  to_torus_dim(r0, r1, 0, 1);
      where (*sp)	/* Block which just finished time step */
       {  CM_u_move_1L(x1_addr, x0_addr, vsz);	/* x1 = x0 */
	  CM_u_move_1L(v1_addr, v0_addr, vsz);	/* v1 = v0 */
	  CM_u_move_1L(d1_addr, d0_addr, vsz);	/* d1 = d0 */
	  r1 = *r0;
       }
      *sp = ( *c == Q0 );	/* Advance sp */
    }	/* End "if" case */
   else	/* Propagate x1, etc. */
    { where ( !*sp )
       { from_torus_dim(x0, x1, vsz, 0, -1);
	 from_torus_dim(v0, v1, vsz, 0, -1);
	 from_torus_dim(d0, d1, vsz, 0, -1);
	 *r0 = from_torus_dim(&r1, 0, -1);  }
    }	/* End "else" case */
   where (*act)   /* Calculate scalar multiplier */
    {  r01 = (*r0 + r1);
       *w =  r01**r0;	   			/* p1 = r1^2 */
       *w += (*w + p1)**r0;  p1 *= r1;		/* p1 = r1^3 */
       *w += (*w + p1)**r0;  p1 *= r1;		/* p1 = r1^4 */
       *w += (*w + p1)**r0;  p1 *= r1;		/* p1 = r1^5 */
       *w += (*w + p1)**r0;  p1 *= r1;		/* p1 = r1^6 */
       *w += p1;	     p1 *= r1;		/* p1 = r1^7 */
       p0 = *r0**r0;  p0 *= *r0*p0*p0;		/* p0 = r0^7 */
       *w = HG**w/(p0*p1);
       print_x("accel repulsion", w, 8, Q0);	/* For testing */
       *w = *mj*( 1.0f/(*r0*r1) - *w )/r01;
       print_x("accel multiplier", w, 8, Q0);	/* For testing */
    }
   for ( u = 0;  u < 3;  u++ )
     { ac[u] = *w*(d0[u]+d1[u]);	/* Acceleration component */
       ac[u] = spread(ac[u], 2, CMC_combiner_add);  /* Sum accelerations */
     }
  }

/* Greenspan's 1st-order method.  Performs ns steps of time dt		*/
/* with solution tolerance eps.  Returns number of evaluations.		*/
/* Leaves number of iterations which were stopped at limit in cfail.	*/
/* All time steps should be initialized with appropriate values for	*/
/*        x0, v0, d0, r0, x1, v1					*/
/* '0' items should be spread over all processors in current time step;	*/
/* sp marks the block for the current time step, in which "0" items are	*/
/* fixed.  Some printing for debugging is provided in comments.		*/
/*******   Must be called with pair shape selected.		*********/


int cfail;   /* Counts number of iterations that failed to converge */

int green1(int ns, float dt, float eps)
 { int k, ic, v, lim=10, u;  float hdt, mx;  float:pair nx, px;
   v = 0;  hdt = 0.5*dt;
   /* print_x("green1, x0 initially", x0, 7, Q0);	For testing */
   /* print_x("green1, v0 initially", v0, 7, Q0);	For testing */
   for (k = 0; k < ns; k++)
     { ic = 0;
       /* This is a do - while - repeat loop: */
       do
        { accel();  v++;  ic++;
	  px = 0.0f;	/* For max change */
	  where (*sm)
	  { for ( u = 0;  u < 3;  u++ )
	     { nx = v0[u] + dt*ac[u];			/* New velocity	*/
	       px >?= fabs(nx - v1[u]);		/* Max change	*/
	       v1[u] = nx;
	       nx = x0[u] + hdt*(v0[u] + v1[u]);	/* New position	*/
	       px >?= fabs(nx - x1[u]);		/* Max change	*/
	       x1[u] = nx; 
	     } /* end for u */
	    where (*sp) mx = (>?= px);   /* Global max in current step*/
  	    /* print_x("green1, v1", v1, 7, Q0);	For testing */
   	    /* print_x("green1, x1", x1, 7, Q0);	For testing */
	    } /* end where (*sm) */
	  if ( (mx < eps) || ( ic == lim ) ) break;
	 }
       while (1);
       advance = 1;	/* Will advance time step in next accel */
       if (ic == lim) cfail++;
     } ;   /* end for */
   return(v);
 }


/* Initialize masses and system state for green1.	*/

void initgreen()
 { int u;  float:pair r1;
   with (pair)
    { /* Spread state to forward blocks */
      CM_spread_with_copy_1L(x0_addr, x0_addr, 0, vsz, Q0);
      CM_spread_with_copy_1L(v0_addr, v0_addr, 0, vsz, Q0);
      *m = copy_spread(m, 0, Q0);
      /* Spread masses for particle j and scale */
      *m = [*c][*j][*j]*m;  *mj = G**m;
      spreadi(d0, x0);	/* For initial displacements */
      /* print_x("initgreen, d0 after spread", d0, 7);	For testing */
      *w = 0.0f; 	/* For initial distances */
      for ( u = 0;  u < 3;  u++ )
       { r1 = d0[u] -= x0[u];  *w += r1*r1; }
      *r0 = sqrt(*w);
      /* print_x("initgreen, d0 with displacements", d0, 7); For testing */
      /* print_x("initgreen, r0", r0, 7);		For testing */
      CM_u_move_1L(x1_addr, x0_addr, vsz);	/* x1 = x0 */
      CM_u_move_1L(v1_addr, v0_addr, vsz);	/* v1 = v0 */
    }  /* End with (pair) */
 }

/* Parameters to control current simulation	*/

int   ns,	/* Steps between output		*/
     npl,	/* Plots per print		*/
      np;	/* Number of prints		*/

float dt,	/* delta-t			*/
      eps;	/* convergence parameter	*/

double tm;	/*current simulation time	*/

/*********       Checkpointing and resuming       *********/

int dvfd;		/* File descriptor for datavault */
FILE *chstrm;		/* File descriptor for host file */

/* Buffer sizes: */

#define ibfsz 7
#define fbfsz 3

/* Checkpoint current calculation using fnm for both host and CM files */

void checkpoint(char *fnm)
{int ibf[ibfsz];  float fbf[fbfsz];
 with (pair)
 { dvfd = CMFS_creat(fnm, 0600);
   CMFS_write_file(dvfd, x0, vsz);  CMFS_write_file(dvfd, v0, vsz);
   CMFS_write_file(dvfd, x1, vsz);  CMFS_write_file(dvfd, v1, vsz);
   CMFS_write_file(dvfd, d0, vsz);  CMFS_write_file(dvfd, r0, vsz);
   CMFS_write_file(dvfd, m,  fsz);  CMFS_write_file(dvfd, mj, fsz);
   CMFS_close(dvfd);
   chstrm = fopen(fnm, "w");
   fbf[0] = dt;  fbf[1] = eps;  fbf[2] = G;
   fwrite(fbf, sizeof(float), fbfsz, chstrm);
   fwrite(&tm, sizeof(double), 1, chstrm);
   ibf[0] = Q;   ibf[1] = N;   ibf[2] = NB;  ibf[3] = advance;
   ibf[4] = ns;  ibf[5] = np;  ibf[6] = Q0;
   fwrite(ibf, sizeof(int), ibfsz, chstrm);
   fclose(chstrm);
  }
 }

/* Resume saved calculation from files fnm (host and CM),	*/
/* initializing if init.  Returns 0 if successful, -1 if not.	*/
/* CM must have proper geometry if not init, should be		*/
/* uninitialized if init.					*/

int resume(char *fnm, int init)
 { float fbf[fbfsz];  int qq, nn, ibf[ibfsz];
   chstrm = fopen(fnm, "r");
   fread(fbf, sizeof(float), fbfsz, chstrm);
   fread(&tm, sizeof(double), 1, chstrm);
   fread(ibf, sizeof(int), ibfsz, chstrm);  fclose(chstrm);
   qq = ibf[0];  nn = ibf[1];
   if ( init )
     { Q = qq;  N= nn;
       allocate_shape(&pair, 3, Q, N, N);
       create_globals(); }
    else if ( (qq != Q) || (nn != N) )
     { printf("Wrong shape - can't resume.\n");  return (-1); }
   dt = fbf[0];  eps = fbf[1];  G = fbf[2];
   NB = ibf[2];  advance = ibf[3];  ns = ibf[4];  np = ibf[5];
   Q0 = ibf[6];  setup();
   with (pair)
    { dvfd = CMFS_open(fnm, CMFS_O_RDONLY);
      CMFS_read_file(dvfd, x0, vsz);  CMFS_read_file(dvfd, v0, vsz);
      CMFS_read_file(dvfd, x1, vsz);  CMFS_read_file(dvfd, v1, vsz);
      CMFS_read_file(dvfd, d0, vsz);  CMFS_read_file(dvfd, r0, vsz);
      CMFS_read_file(dvfd, m,  fsz);  CMFS_read_file(dvfd, mj, fsz);
      CMFS_close(dvfd);
    }
   return (0);
 }

/* Initialize problem from data file */

void fileinit(char *fnm)
 { float ar, am;
   datap = fopen(fnm, "r");
   fscanf(datap, "%d", &NB);
   setup();   /* Setup geometric structure */
   printf("Initialize....\n");
   init();    /* Read data for individual bodies */
   /* Read "asteroid" descriptions */
   while ( fscanf(datap, "%d", &ns), ns > 0)
     { fscanf(datap, "%e%e", &ar, &am);  mkast(ns, ar, am); }
   fscanf(datap, "%d%d%e%e", &ns, &np, &dt, &eps);
   fclose(datap);
   tm = 0.0;
   initgreen();	/* Set up for integration */
 }

/* Get projection parameters from user and update */

void plotparams()
 { float cx, cy, xw;
   printf("Enter plots/phase, drgb, center-x, center-y, width > ");
   scanf("%d%d%e%e%e", &npl, &drgb, &cx, &cy, &xw);
   setup_proj(cx, cy, xw);
 }

/* Initialize display for plotting */

void initplot()
 { finddisp();			/* Attach and initialize */
   with (pair)
    { load_solar(i, j, N);	/* Set up icons */
      plotparams(); }		/* Get parameters */
 }


/* Main program:  reads problem and control information from data file	*/
/* named by user and performs calculation.				*/
/* Print control:							*/
/*	 -2 - plot bodies (requests additional info)			*/
/*	  0 - just performance data					*/
/*	 -1 - all							*/
/*	> 0 - number given						*/
/* Commands:								*/
/*	  c - continue calculation					*/
/*	  x - checkpoint						*/
/*	  r - resume							*/
/*	  p - modify parameters						*/
/*	  q - quit							*/

void main()
 { int k, j, vl, tvl, tfail, px;
   double eltm, cmtm, pr;	/* Elapsed, CM times, misc. */
   char fnm[50], dgp[50] ;	/* dgp for control   */
   /*  \007 = BEL, works in either old or new C */
   printf("\007Diagnostic printing (y or n)\007? ");
   fgets(dgp, sizeof(dgp), stdin);
   PD = ( dgp[0] == 'y' ) ? 1 : 0;
   printf("Asteroid interaction radius (default = %f)? ", Rx);
   fgets(dgp, sizeof(dgp), stdin); 
   if ( dgp[0] != 0 ) sscanf(dgp, "%f", &Rx);
   printf("Initialize (i) or resume (r)? > ");
   scanf("%s", dgp);
   if ( dgp[0] == 'r' )
     { printf("Checkpoint file> ");  scanf("%s", fnm);
       resume(fnm, 1); }
   else   /* Initialize the pair shape with suitable geometry */
     { printf("Enter Q, N for [Q][N][N] shape> ");
       scanf("%d%d", &Q, &N);
       allocate_shape(&pair, 3, Q, N, N);
       create_globals();
       printf("Data file> ");  scanf("%s", fnm);
       fileinit(fnm); }
   pr = Rx*Rx;  HG = pr*pr*pr;  H = G*HG;	/* Repulsive constant */
   printf("Print ctl (-2 for plot)> ");  scanf("%d", &PX);
   if ( PX == -2 ) initplot(); else npl = 1;
   printf("Shape is [%d][%d][%d] = %d vp / %d physical\n",
                     Q,  N,  N, Q*N*N, positionsof(physical) );
   printf("\n   NB = %d  G = %13.6e  H = %13.6e",
		     NB,        G,          H);
   printf("\n   delta-t = %12.5e  epsilon = %12.5e\n",
                            dt,               eps);
   printf("\n simulated time = %9.3f", tm);
   if ( PX == -2 ) printf("\n");  with (pair) print_state();
   /* Loop for simulation runs */
   while ( printf("c, x, r, p, or q)> "),
	   scanf("%s", fnm),
	   ( fnm[0] != 'q' ) || ( fnm[1] > 0 ) )
   if ( ( fnm[0] == 'x' ) && ( fnm[1] == 0 ) )		/* Checkpoint */
     { printf("Checkpoint file> ");  scanf("%s", fnm);
       checkpoint(fnm); }
  else if ( ( fnm[0] == 'r' ) && ( fnm[1] == 0 ) )	/* Resume */
     { printf("Checkpoint file> ");  scanf("%s", fnm);
       resume(fnm, 0);  /* Geometry already set up */ }
  else if ( ( fnm[0] == 'p' ) && ( fnm[1] == 0 ) )	/* Parameters */
     { printf("Enter PX, ns, np, dt, eps > ");
       scanf("%d%d%d%e%e", &px, &ns, &np, &dt, &eps);
       if ( px == -2 && PX != -2 ) initplot();
       else if ( px != -2 && PX == -2 )	/* Change to printing */
	{ detach();   npl = 1; }
       if ( PX == -2 ) plotparams();
       PX = px; }
  else  /* Continue simulation */
    { tvl = 0;  tfail = 0;  CM_timer_start(1);
      with (pair)
      for ( k = 0; k < np; k++ )
       { cfail = 0;  vl = 0;
         for ( j = 0; j < npl; j++ )
	   { vl += green1(ns, dt, eps);
	     if ( PX == -2 ) print_state();
           }  /* end for j */
	 tm += ns*npl*dt;  tvl += vl;  tfail += cfail;
         printf(
     "\n simulated time = %9.3f  %d evaluations  %d iterations limited\n",
		            tm,  vl,            cfail);
	 if ( PX > -2 ) print_state();
       }  /* end for k */
      CM_timer_stop(1);
      eltm = CM_timer_read_elapsed(1);  cmtm = CM_timer_read_cm_busy(1);
      printf(
     "\n simulated time = %9.3f  %d evaluations  %d iterations limited\n",
		            tm,  tvl,           tfail);
      printf(
       "Elapsed: %6.3f sec   CM: %6.3f sec   %3.0f%% utilization\n",
     	       eltm,           cmtm,  100.0*(cmtm/eltm) );
      eltm = 1000*eltm;  cmtm = 1000*cmtm;	/* Now in msec. */
      j = np*npl*ns;	/* Total number of time steps */
      printf("%3.1f : %3.1f msec/eval,  %3.1f : %3.1f msec/step\n",
     	    eltm/tvl, cmtm/tvl,	        eltm/j, cmtm/j );
      CM_timer_clear(1);
      printf("\007");		/* BEL */
    }  /* end continue simulation */
 }

/*********************************************************************
  Data files:  as for nb3dg2.cs
*************************************************************************/
